from transformers.optimization import AdamW
from ratsnlp.nlpbook.metrics import accuracy
from pytorch_lightning import LightningModule
from transformers import BertPreTrainedModel
from ratsnlp.nlpbook.ner import NERTrainArguments, NER_PAD_ID
from pytorch_lightning.trainer.supporters import TensorRunningAccum
from torch.optim.lr_scheduler import ExponentialLR, CosineAnnealingWarmRestarts


class NERTask(LightningModule):

    def __init__(self,
                 model: BertPreTrainedModel,
                 args: NERTrainArguments,
    ):
        super().__init__()
        self.model = model
        self.args = args
        self.running_accuracy = TensorRunningAccum(window_length=args.stat_window_length)

    def configure_optimizers(self):
        if self.args.optimizer == 'AdamW':
            optimizer = AdamW(self.parameters(), lr=self.args.learning_rate)
        else:
            raise NotImplementedError('Only AdamW is Supported!')
        if self.args.lr_scheduler == 'cos':
            scheduler = CosineAnnealingWarmRestarts(optimizer, T_0=1, T_mult=2)
        elif self.args.lr_scheduler == 'exp':
            scheduler = ExponentialLR(optimizer, gamma=0.5)
        else:
            raise NotImplementedError('Only cos and exp lr scheduler is Supported!')
        return {
            'optimizer': optimizer,
            'scheduler': scheduler,
        }

    def forward(self, **kwargs):
        return self.model(**kwargs)

    def step(self, inputs, mode="train"):
        loss, logits = self.model(**inputs)
        preds = logits.argmax(dim=-1)
        labels = inputs["labels"]
        acc = accuracy(preds, labels, ignore_index=NER_PAD_ID)
        self.running_accuracy.append(acc)
        logs = {f"{mode}_loss": loss, f"{mode}_acc": acc}
        return {"loss": loss, "log": logs}

    def training_step(self, inputs, batch_idx):
        return self.step(inputs, mode="train")

    def validation_step(self, inputs, batch_idx):
        return self.step(inputs, mode="val")

    def test_step(self, inputs, batch_idx):
        return self.step(inputs, mode="test")

    def epoch_end(self, outputs, mode="train"):
        loss_mean, acc_mean = 0, 0
        for output in outputs:
            loss_mean += output['loss']
            acc_mean += output['log'][f'{mode}_acc']
        acc_mean /= len(outputs)
        results = {
            'log': {
                f'{mode}_loss': loss_mean,
                f'{mode}_acc': acc_mean,
            },
            'progress_bar': {f'{mode}_acc': acc_mean},
        }
        return results

    def validation_epoch_end(self, outputs):
        return self.epoch_end(outputs, mode="val")

    def test_epoch_end(self, outputs):
        return self.epoch_end(outputs, mode="test")

    def get_progress_bar_dict(self):
        running_train_loss = self.trainer.running_loss.mean()
        running_train_accuracy = self.running_accuracy.mean()
        tqdm_dict = {
            'tr_loss': '{:.3f}'.format(running_train_loss.cpu().item()),
            'tr_acc': '{:.3f}'.format(running_train_accuracy.cpu().item()),
        }
        return tqdm_dict



import torch
import torch.nn.functional as F
from transformers import PreTrainedModel


def _cosine_scores(questions: torch.Tensor, answers: torch.Tensor, eps: float = 1e-8) -> torch.Tensor:
    """
    questions: n1 x D, answers: n2 x D, result n1 x n2
    pairwise cosine similarity matrix 계산
    이 행렬의 대각성분은 각각 q0-a0, q1-a1, ... 사이의 코사인 유사도가 된다
    이는 F.cosine_similarity(questions, answers, dim=1)와 일치한다
    """
    questions_l2_norm = questions.norm(p=2, dim=1, keepdim=True)
    answers_l2_norm = answers.norm(p=2, dim=1, keepdim=True)
    numerator = torch.matmul(questions, torch.transpose(answers, 0, 1))
    denominator = (questions_l2_norm * torch.transpose(answers_l2_norm, 0, 1)).clamp(min=eps)
    return numerator / denominator


class SearchModelForTrain(torch.nn.Module):

    def __init__(
            self,
            question_tower: PreTrainedModel,
            passage_tower: PreTrainedModel,
            score_fn=_cosine_scores,
    ):
        super().__init__()
        self.question_tower = question_tower
        self.passage_tower = passage_tower
        self.score_fn = score_fn

    def forward(
            self,
            question_input_ids=None,
            question_attention_mask=None,
            question_token_type_ids=None,
            passage_input_ids=None,
            passage_attention_mask=None,
            passage_token_type_ids=None,
            labels=None,
    ):
        question_embeddings = self.question_tower(
            input_ids=question_input_ids,
            token_type_ids=question_token_type_ids,
            attention_mask=question_attention_mask,
        )[1]
        passage_embeddings = self.passage_tower(
            input_ids=passage_input_ids,
            token_type_ids=passage_attention_mask,
            attention_mask=passage_token_type_ids,
        )[1]
        scores = self.score_fn(question_embeddings, passage_embeddings)
        softmax_scores = F.log_softmax(scores, dim=1)
        loss = F.nll_loss(softmax_scores, labels, reduction='mean')
        return loss, softmax_scores


class SearchModelForInference(torch.nn.Module):

    def __init__(
            self,
            question_tower: PreTrainedModel,
            passage_tower: PreTrainedModel,
            score_fn=_cosine_scores,
    ):
        super().__init__()
        self.question_tower = question_tower
        self.passage_tower = passage_tower
        self.score_fn = score_fn

    def forward(
            self,
            question_input_ids=None,
            question_attention_mask=None,
            question_token_type_ids=None,
            passage_input_ids=None,
            passage_attention_mask=None,
            passage_token_type_ids=None,
            passage_embeddings=None,
            mode="inference",
    ):
        if mode == "inference":
            question_embeddings = self.question_tower(
                input_ids=question_input_ids,
                token_type_ids=question_token_type_ids,
                attention_mask=question_attention_mask,
            )[1]
            scores = self.score_fn(question_embeddings, passage_embeddings)
            return scores
        else:
            passage_embeddings = self.passage_tower(
                input_ids=passage_input_ids,
                token_type_ids=passage_attention_mask,
                attention_mask=passage_token_type_ids,
            )[1]
            return passage_embeddings



from .arguments import SearchTrainArguments, SearchDeployArguments
from .task import SearchTask
from .corpus import KorQuADV1Corpus, SearchDataset, search_train_collator
from .deploy import get_web_service_app, encoding_passage
from .modeling import SearchModelForTrain, SearchModelForInference



from tokenizers import SentencePieceBPETokenizer
from transformers.tokenization_utils_fast import PreTrainedTokenizerFast


MASK_TOKEN = "<unused0>"


class KoGPT2Tokenizer(PreTrainedTokenizerFast):
    """
    reference:
        taeminlee/KoGPT-Transformers
        https://github.com/taeminlee/KoGPT2-Transformers/blob/master/kogpt2_transformers/kogpt2_transformers/tokenization_kogpt2.py
    """

    vocab_files_names = {
        "vocab_file": "vocab.json",
        "merges_file": "merges.txt",
        "special_tokens_map_file": "special_tokens_map.json",
        "added_tokens_file": "added_tokens.json"
    }

    def __init__(
        self,
        vocab_file,
        merges_file,
        unk_token="<unk>",
        bos_token="<s>",
        eos_token="</s>",
        pad_token="<pad>",
        add_prefix_space=False,
        **kwargs
    ):
        super().__init__(
            SentencePieceBPETokenizer(
                vocab_file=vocab_file,
                merges_file=merges_file,
                add_prefix_space=add_prefix_space,
            ),
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            pad_token=pad_token,
            **kwargs,
        )

    @classmethod
    def from_pretrained(cls, *inputs, **kwargs):
        return cls._from_pretrained(*inputs, **kwargs)



import os
import re
import json
import time
import torch
import random
import logging
from tqdm import tqdm
from filelock import FileLock
from dataclasses import dataclass
from typing import List, Optional
from collections import defaultdict
from torch.utils.data.dataset import Dataset
from transformers import PreTrainedTokenizer
from ratsnlp.nlpbook.search.arguments import SearchTrainArguments

logger = logging.getLogger(__name__)


@dataclass
class SearchExample:
    question: str
    passage: str
    group: Optional[int] = None


@dataclass
class SearchFeatures:
    input_ids: List[int]
    attention_mask: Optional[List[int]] = None
    token_type_ids: Optional[List[int]] = None


@dataclass
class SearchPositivePair:
    question_features: SearchFeatures
    passage_features: SearchFeatures


class SearchCorpus:

    def __init__(self):
        pass

    def get_examples(self, corpus_dir):
        """
        :return: List[SearchExample]
        """
        raise NotImplementedError


class KorQuADV1Corpus(SearchCorpus):

    def __init__(self):
        super().__init__()
        self.train_file = "KorQuAD_v1.0_train.json"
        self.val_file = "KorQuAD_v1.0_dev.json"

    def get_examples(self, corpus_dir):
        examples = []
        train_corpus_fpath = os.path.join(corpus_dir, self.train_file)
        val_corpus_fpath = os.path.join(corpus_dir, self.val_file)
        json_train_data = json.load(open(train_corpus_fpath, "r", encoding="utf-8"))["data"]
        json_val_data = json.load(open(val_corpus_fpath, "r", encoding="utf-8"))["data"]
        json_data = json_train_data + json_val_data
        group_id, num_elements = 0, 0
        for entry in tqdm(json_data):
            title = re.sub(r'_|-|\(|\)|~', ' ', entry["title"]).strip()
            for paragraph in entry["paragraphs"]:
                context_text = paragraph["context"]
                for qa in paragraph["qas"]:
                    question_text = qa["question"]
                    if title and question_text and context_text:
                        example = SearchExample(
                            question=title,
                            passage=context_text,
                            group=group_id,
                        )
                        examples.append(example)
                        example = SearchExample(
                            question=question_text,
                            passage=context_text,
                            group=group_id,
                        )
                        examples.append(example)
                        num_elements += 2
                # context마다 그룹ID 부여
                if num_elements > 0:
                    group_id += 1
                    num_elements = 0
        return examples


def _convert_examples_to_search_features(
        examples: List[SearchExample],
        tokenizer: PreTrainedTokenizer,
        args: SearchTrainArguments,
):

    logger.info(
        "tokenize sentences, it could take a lot of time..."
    )
    start = time.time()
    question_batch_encoding = tokenizer(
        [example.question for example in examples],
        max_length=args.question_max_seq_length,
        padding="max_length",
        truncation=True,
    )
    passage_batch_encoding = tokenizer(
        [example.passage for example in examples],
        max_length=args.passage_max_seq_length,
        padding="max_length",
        truncation=True,
    )
    logger.info(
        "tokenize sentences [took %.3f s]", time.time() - start
    )

    features = defaultdict(list)
    for i in range(len(examples)):
        question_inputs = {k: question_batch_encoding[k][i] for k in question_batch_encoding}
        question_features = SearchFeatures(**question_inputs)
        passage_inputs = {k: passage_batch_encoding[k][i] for k in passage_batch_encoding}
        passage_features = SearchFeatures(**passage_inputs)
        features[examples[i].group].append([question_features, passage_features])

    for group_id in list(features.keys())[:5]:
        logger.info("*** Example group id: %s ***" % (group_id))
        for all_features in features[group_id]:
            q_features, p_features = all_features
            question_tokens = tokenizer.convert_ids_to_tokens(q_features.input_ids)
            passage_tokens = tokenizer.convert_ids_to_tokens(p_features.input_ids)
            positive_pair = " ".join([t for t in question_tokens if t != tokenizer.pad_token]) + "/" \
                            + " ".join([t for t in passage_tokens if t != tokenizer.pad_token])
            logger.info("positive pair tokens: %s" % (positive_pair))
            logger.info("question features: %s" % q_features)
            logger.info("passage features: %s" % p_features)

    return features


def search_train_collator(positive_pairs):
    features = defaultdict(list)
    for positive_pair in positive_pairs:
        for q_k, q_v in vars(positive_pair.question_features).items():
            features[f"question_{q_k}"].append(q_v)
        for p_k, p_v in vars(positive_pair.passage_features).items():
            features[f"passage_{p_k}"].append(p_v)
    labels = [el for el in range(len(positive_pairs))]
    all_features = {k: torch.LongTensor(v) for k, v in {**features, "labels": labels}.items()}
    return all_features


class SearchDataset(Dataset):

    def __init__(
            self,
            args: SearchTrainArguments,
            tokenizer: PreTrainedTokenizer,
            corpus: SearchCorpus,
            mode: Optional[str] = "train",
            convert_examples_to_features_fn=_convert_examples_to_search_features,
    ):
        if corpus is not None:
            self.corpus = corpus
        else:
            raise KeyError("corpus is not valid")
        if not mode in ["train", "inference"]:
            raise KeyError(f"mode({mode}) is not a valid split name")
        else:
            self.mode = mode

        # Load data features from cache or dataset file
        cached_features_file = os.path.join(
            args.downstream_corpus_root_dir,
            args.downstream_corpus_name,
            "cached_{}_query{}_passage{}_{}_{}".format(
                tokenizer.__class__.__name__,
                str(args.question_max_seq_length),
                str(args.passage_max_seq_length),
                args.downstream_corpus_name,
                "document-search",
            ),
        )

        # Make sure only the first process in distributed training processes the dataset,
        # and the others will use the cache.
        lock_path = cached_features_file + ".lock"
        with FileLock(lock_path):

            if os.path.exists(cached_features_file) and not args.overwrite_cache:
                start = time.time()
                features = torch.load(cached_features_file)
                logger.info(
                    f"Loading features from cached file {cached_features_file} [took %.3f s]", time.time() - start
                )
            else:
                corpus_dir = os.path.join(args.downstream_corpus_root_dir, args.downstream_corpus_name)
                logger.info(f"Creating features from dataset file at {corpus_dir}")
                examples = self.corpus.get_examples(corpus_dir)
                features = convert_examples_to_features_fn(
                    examples,
                    tokenizer,
                    args,
                )
                start = time.time()
                logging.info(
                    "Saving features into cached file, it could take a lot of time..."
                )
                torch.save(features, cached_features_file)
                logger.info(
                    "Saving features into cached file %s [took %.3f s]", cached_features_file, time.time() - start
                )

            if self.mode == "train":
                self.features = features
            else:
                passages = set()
                passage_features = []
                for features_in_a_group in features.values():
                    for el in features_in_a_group:
                        _, passage_feature = el
                        passage = " ".join([str(el) for el in passage_feature.input_ids])
                        if passage not in passages:
                            passage_features.append(passage_feature)
                            passages.add(passage)
                self.features = passage_features

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i):
        if self.mode == "train":
            # i번째 그룹에서 포지티브 페어를 랜덤으로 하나 뽑는다
            all_features = random.sample(self.features[i], 1)[0]
            question_features, passage_features = all_features
            return SearchPositivePair(
                question_features=question_features,
                passage_features=passage_features,
            )
        else:
            # inference
            return self.features[i]



import os
import time
import torch
import logging
from tqdm import tqdm
from flask import Flask, request, jsonify, render_template

logger = logging.getLogger(__name__)


def encoding_passage(inference_dataloader, model, tokenizer, args):
    passage_embedding_fpath = os.path.join(
            args.passage_embedding_dir,
            "passage-embeddings_{}_{}".format(
                tokenizer.__class__.__name__,
                str(args.passage_max_seq_length),
            ),
        )
    if os.path.exists(passage_embedding_fpath) and not args.overwrite_cache:
        start = time.time()
        all_passages, all_passage_embeddings = torch.load(passage_embedding_fpath, map_location=torch.device("cpu"))
        logger.info(
            f"Loading passage embeddings from cached file {passage_embedding_fpath} [took %.3f s]", time.time() - start
        )
    else:
        logger.info(f"Creating passage embeddings ...")
        all_passages = []
        all_passage_embeddings = []
        special_tokens = [tokenizer.cls_token_id, tokenizer.sep_token_id, tokenizer.pad_token_id]
        with torch.no_grad():
            for batch_inputs in tqdm(inference_dataloader, desc="passage encoding"):
                batch_input_ids = batch_inputs["input_ids"]
                for input_ids in batch_input_ids:
                    input_ids = [el for el in input_ids if el not in special_tokens]
                    all_passages.append(tokenizer.decode(input_ids))
                if torch.cuda.is_available():
                    batch_inputs = {k: v.cuda() for k, v in batch_inputs.items()}
                # passage_embeddings : args.batch_size x hidden_dimension
                passage_embeddings = model(**{**{f"passage_{k}": v for k, v in batch_inputs.items()}, "mode": "encoding"})
                all_passage_embeddings.append(passage_embeddings)
            # all passage_embeddings : total_num_of_passage x hidden_dimension
            all_passage_embeddings = torch.cat(all_passage_embeddings)
        start = time.time()
        logging.info(
            "Saving passage embeddings into cached file, it could take a lot of time..."
        )
        torch.save((all_passages, all_passage_embeddings), passage_embedding_fpath)
        logger.info(
            "Saving passage embeddings into cached file %s [took %.3f s]", passage_embedding_fpath, time.time() - start
        )
    return all_passages, all_passage_embeddings


def get_web_service_app(inference_fn, is_colab=True):

    app = Flask(__name__, template_folder='')
    if is_colab:
        from flask_ngrok import run_with_ngrok
        run_with_ngrok(app)
    else:
        from flask_cors import CORS
        CORS(app)

    @app.route('/')
    def index():
        return render_template('index.html')

    @app.route('/api', methods=['POST'])
    def api():
        query = request.json
        output_data = inference_fn(query["question"])
        response = jsonify(output_data)
        return response

    return app
